/*
  pico gps.c  Use GPS altitude for comparison with Pressure Altitude 
  Jed Margolin 1/162025
  board: pico-gps-r4   Can be powered by +9V or by USB power, the choice is made when the board is stuffed
  Use 4-pin Blue NEO7M GPS module
  Use bme280 for temperature, humidity, and air pressure
  
  Logs the data in Flash: Local Time, GPS Coordinates, GPS Altitude, Temperature, Pressure, 
  Pressure Altitude, Humidity, and Barometric Reference. 
  
  Turn off all GPS messages, then turn on the ones we want: GGA, ZDA, PUBX0, and PUBX4.
  uart is 115,200 Baud, 1-start, 8 data, 1 stop, no parity.

  pref_work.pref_bcd is the barometric reference in bcd, and keeps it in other formats.

  When SW is on Settings it goes to the lcd screen to set barometric reference, local hour offset from UTC, and 
  whether DST is observed.

  If PB1 is pressed during startup it will echo the GPS uart characters to the screen. 
  It needs to be pressed until the test starts.
  After it starts, pressing PB3 will stall it as long as it is pressed.
  This is a basic test, and does not come back.

  If PB2 is pressed during startup it will echo the GPS messages that have been received to the screen. 
  It needs to be pressed until the test starts.
  After it starts, pressing PB3 will stall it as long as it is pressed.
  This tests that the GPS messages are being detected, does not come back

  Once the main program starts, pressing PB1 will show just #sats and Leap Seconds while PB1 is pressed.
  
  uart0 can use the ch340 (uart to USB adapter) with the AccessPort program for Windows (https://accessport.soft32.com).

  The 'M' in NEO7M means its firmware is in masked ROM. Thus, it cannot be updated.

  The NEO7M does not support external EEPROM. (The NEO6M did support it.) If your board has one, it is not being used. 
  Data can be saved only in the NEO7M's onchip RAM and only if there is sufficient backup voltage. With the NEO7M boards 
  from China this backup voltage will probably come from a super capacitor and will not last very long. With one of my 
  boards it lasted 30 minutes. With another board it lasted 30 seconds. If you want to save data add your own backup battery. 
  My pico-gps-r4 board allows you to do this. This requires soldering a wire to the bracket on the top of the super capacitor 
  on the NEO7M module and connecting it to a pin on the board. You can use a CR2032 button cell or two AAA batteries in series. It will save data for almost the 
  shelf life of the batteries. 

  Without this, the NEO7M will have to do a cold start every time power is applied. It can take up to 12.5 minutes to acquire
  the correct Leap Seconds after sufficient sats are acquired.   :-(

  Once it starts in normal mode there are commands you can issue using the keyboard. 
  This includes dumping the Flash Data Logs.
*/

#include <stdio.h>
#include <string.h>
#include "pico/stdlib.h"
#include "hardware/i2c.h"
#include "hardware/uart.h"
#include "hardware/structs/i2c.h"
#include "pico/binary_info.h"
#include <math.h>
#include "main.h"          // mine
#include "gps_control.h"   // mine

#define FLAG1 0     // For testing
//================================================================
// Get all of the GPS messages from '$' to 0x0A (CR) regardless of the message
// Use irqs for reading GPS messages, use RX FIFO
void on_uart_rx1() 
{
   while(uart_is_readable(UART_ID1)) 
   {
      rxdata = uart_getc(UART_ID1);

      if(rxdata == '$')  
      {
          gps_data_count = 0; 
          gps_go = 1;
      }

      if(gps_go)
      {
         gps_buffer[gps_data_count++] = rxdata;

         if(gps_data_count > 250)  // No message should be this long, something went wrong, start over
         {
            gps_go = 0;
            gps_data_count = 0;
            return;
         }
 
         if(rxdata == 0x0A)  
         {
            gps_buffer[gps_data_count] = 0;    // NULL to terminate the string
            gps_flag = 1;
            gps_go = 0;
            gps_data_count = 0;

            // We have a new message. What is it?
            strncpy(comp_str,gps_buffer,6);    // Get message name

//           if( (strcmp(zda_str,comp_str) == 0 ) )
             if( (strcmp(zda_str,comp_str) == 0 ) && (!zda_flag) ) // We have a match to $GPZDA and are ready for a new ZDA message
            { 
               strcpy(zda_buffer,gps_buffer);
//             uart_puts(UART_ID, zda_buffer);
               zda_flag = 1;
            }

//          if( (strcmp(gga_str,comp_str) == 0 ) )
            if( (strcmp(gga_str,comp_str) == 0 ) && (!gga_flag) )  // We have a match to $GPGGA and are ready for a new GGA message
            { 
               strcpy(gga_buffer,gps_buffer);
//             uart_puts(UART_ID, gga_buffer);
               gga_flag = 1;
            }

//          if(strcmp(pubx_str,comp_str) == 0 )           
            if( (strcmp(pubx_str,comp_str) == 0 ) && (!pubx_flag) ) // We have a match to $PUBX, and are ready for a new PUBX message
            { 
               strcpy(pubx_buffer,gps_buffer);
//             uart_puts(UART_ID, pubx_buffer);
               pubx_flag = 1;
            }
         
         } // end gps_go     
     
      }  // end of message

   }  // end - while uart is readable

    return;
}
//================================================================
// Used to read and debounce the switches during setttings mode
bool repeating_timer_callback(struct repeating_timer *t) 
{
   irq_flag = 1;

   // read switches
   do_switch(PB1, &switch1);   // PB1, switch structure  INC
   do_switch(PB2, &switch2);   // PB2, switch structure  DEC
   do_switch(PB3, &switch3);   // PB3, switch structure  MODE
   
   irq_count--;
   if(irq_count == 0)  
   { 
      irq_count = 250;   // 250 = 1 second
      secs_flag = 1;      // seconds flag
   }

   return (true);
}
//===============================================================
int main () 
{
   uint16_t n;

// Hardware on the board
   gpio_init(LED1);      // Green
   gpio_set_dir(LED1, GPIO_OUT);
   gpio_put(LED1,0);

   gpio_init(LED2);      // Red
   gpio_set_dir(LED2, GPIO_OUT);
   gpio_put(LED2,0);

   gpio_init(PB1);       // Pushbutton 1
   gpio_set_dir(PB1, GPIO_IN);
   gpio_pull_up(PB1);

   gpio_init(PB2);       // Pushbutton 2
   gpio_set_dir(PB2, GPIO_IN);
   gpio_pull_up(PB2);

   gpio_init(PB3);       // Pushbutton 3
   gpio_set_dir(PB3, GPIO_IN);
   gpio_pull_up(PB3);

   gpio_init(SW);        // Slide Switch
   gpio_set_dir(SW, GPIO_IN);
   gpio_pull_up(SW);

   gpio_init(CHAN);      // Header Pin, make it an output for testing
   gpio_set_dir(CHAN, GPIO_OUT);
   gpio_put(CHAN,0);
//-----------------------------------
// For the bme-280
// This uses I2C0 on the default SDA and SCL pins (4, 5)
   i2c_init(i2c_default, 300 * 1000);   // 300 KHz (400 KHz nax)
   gpio_set_function(PICO_DEFAULT_I2C_SDA_PIN, GPIO_FUNC_I2C);
   gpio_set_function(PICO_DEFAULT_I2C_SCL_PIN, GPIO_FUNC_I2C);
   gpio_pull_up(PICO_DEFAULT_I2C_SDA_PIN);
   gpio_pull_up(PICO_DEFAULT_I2C_SCL_PIN);
//    bi_decl(bi_2pins_with_func(PICO_DEFAULT_I2C_SDA_PIN, PICO_DEFAULT_I2C_SCL_PIN, GPIO_FUNC_I2C));
//-----------------------------------
// Set up the UARTs with the required speed.
   uart_init(UART_ID, BAUD_RATE);

// Set the TX and RX pins by using the function select on the GPIO
   gpio_set_function(UART_TX_PIN, GPIO_FUNC_UART);
   gpio_set_function(UART_RX_PIN, GPIO_FUNC_UART);

// Send the title message and version out through uart
   uart_puts(UART_ID, version);   // send it out
   
// uart for GPS
// Set up the UART with the required speed.
   uart_init(UART_ID1, BAUD_RATE1);

// Set the TX and RX pins by using the function select on the GPIO
   gpio_set_function(UART_TX1_PIN, GPIO_FUNC_UART);
   gpio_set_function(UART_RX1_PIN, GPIO_FUNC_UART);
//----------------------------
   lcd_init();

// Greetings to LCD  - Canned Messages in Flash
   domsg(&dmsg1);    // "Jed Margolin's"
   domsg(&dmsg2);    // "GPS"

   sleep_ms(2000);   // 2 seconds

   SetCmdMode();     // set for commands
   lcd_byte(0x01);   // clear lcd, both lines

   domsg(&dmsg3);    // "[date]"
//------------------------------------
   strcpy(zda_str,"$GPZDA");   // To identify the GPS message
   strcpy(gga_str,"$GPGGA");
   strcpy(pubx_str,"$PUBX,");  // Has to have 6 characters
//------------------------------------
   blink_led(300, 1);
//------------------------------------
   zda_flag = 0;     // 1 = when we have a zda message that we haven't processed yet
   gga_flag = 0;     // 1 = when we have a gga message that we haven't processed yet
   pubx_flag = 0;    // 1 = when we have a pubx message that we haven't processed yet
   gps_flag = 0;     // 1 = message used for testing
   gps_go = 0;
   gps_data_count = 0;
//============================================
// Set up an RX interrupt for uart1 (for GPS)
// We need to set up the handler first
// Select correct interrupt for the UART we are using
   int UART_IRQ1 = UART_ID1 == uart0 ? UART0_IRQ : UART1_IRQ;   // if a then b else c or a ? b : c

// Set up and enable the interrupt handlers
   irq_set_exclusive_handler(UART_IRQ1, on_uart_rx1);
   irq_set_enabled(UART_IRQ1, true);

// Now enable UART1 to send interrupts - RX only
   uart_set_irq_enables(UART_ID1, true, false);
//---------------------------
// uart0: make sure uart FIFO is empty, disable FIFO for this (keyboard and monitor)
   uart_set_fifo_enabled(UART_ID, false);
   for(n=0; n<64; n++)   
   {
       // keep going until FIFO is empty (0 bytes in FIFO)
       if( uart_is_readable(UART_ID) )   { uart_char = uart_getc(UART_ID); }
       else {break;}
   }
//   uart_set_fifo_enabled(UART_ID, true);    // do not use FIFO

// uart1: make sure uart FIFO is empty, Use FIFO for this (GPS module)
   uart_set_fifo_enabled(UART_ID1, false);
   for(n=0; n<64; n++)   
   {
       // keep going until FIFO is empty (0 bytes in FIFO)
       if( uart_is_readable(UART_ID1) )   { uart_char = uart_getc(UART_ID1); }
       else {break;}
   }
   uart_set_fifo_enabled(UART_ID1, true);  // uart FIFO for GPS

   // Does this work? It should cause the RX FIFO to trigger an interrupt when the FIFO has 28 bytes
   uart_get_hw(UART_ID1)->ifls=0x04;  
//=========================================
// neo7m: Disable all GPS Messages, then enable the ones we want

   for(n=0; n<Dis_All_Size; n++) { uart_putc(UART_ID1, Disable_All[n]); }
   sleep_ms(10);  // probably not needed

   for(n=0; n<En_ZDA_Size; n++)  { uart_putc(UART_ID1, Enable_ZDA[n]); }        // Time (corrected UTC)
   for(n=0; n<En_GGA_Size; n++)  { uart_putc(UART_ID1, Enable_GGA[n]); }        // Coordinates and #sats
   for(n=0; n<En_PUBX4_Size; n++)  { uart_putc(UART_ID1, Enable_PUBX4[n]); }    // Leap Seconds
   for(n=0; n<En_PUBX0_Size; n++)  { uart_putc(UART_ID1, Enable_PUBX0[n]); }    // Uncorrected UTC and ellipsoid altitude
//------------------------------------
   if( !gpio_get(PB1) )  { echo_gps_msgs(); }  // Just echo the GPS characters for testing, does not come back
   test_flag = !gpio_get(PB2);   // test that it is receiving messages (Will happen a little later)
//===============================
   irq_flag = 0;
   irq_count = 250;
   secs_flag = 0;
   data_count = 0;

   mode = gpio_get(SW);
   if(!mode)  // Set the barometric reference, local offset, and dst enable
   { 
      // Create a repeating timer that calls repeating_timer_callback
      // If the delay is negative the next callback will be starting from the beginning of the callback,
      // Positive starts from the end of the callback
      add_repeating_timer_ms(-4,repeating_timer_callback, NULL, &timer);  // 4 ms
      set_reference();  // Set Barometer Reference, Local UTC offset, DST enable
   }  
//---------------------------------------------
//   test_expo();  // test exponentiation function
//----------------------------
   blink_led(500, 2);
//---------------------------
   start_bme280();  // Get Device ID and a reading

   if(test_flag)  { test_messages(); }   // test the gps messages, does not come back
//==============================================
// GPS main
   gps_flag = 0;

   strcpy(sats_in_view,"00");
   wait_sats();  // Wait for at least 5 sats in view, also get leap seconds

   sleep_ms(3000);

   SetCmdMode();     // set for commands
   lcd_byte(0x01);   // clear lcd, both lines

   domsg(&dmsg14);   // Wait for ZDA
   
   while(!zda_flag);
   show_zda_buffer(); 
   decode_ZDA(1);
   display_utc_lcd();
   zda_flag = 0;

   sleep_ms(3000);

   SetCmdMode();     // set for commands
   lcd_byte(0x01);   // clear lcd, both lines
//-------------------------
// Get Flash memory ready
   find_first_blank_page();  // Find the first blank Page_Index (and Sector_index)
   sleep_ms(100);

// If the SDATE page is blank, write today as the Start Date, otherwise just show the Start Date
   do_start_date(0);   // 1 = erase and do new start date of the program
   sleep_ms(100);

// If the BREF page is blank: use standard reference of 0x2992 (29.92 inches HG), 
   do_B_reference(0x2992);

// If the LOCAL page is blank: use -8 hours and DST on, otherwise get local_offset and dst_flag 
   do_local(-8, 1, 0);    // default local offset, default dst On, erase

// if dst flash is blank use this year and get dst start and end dates
   do_dst(0);      // 1 = erase first
//--------------------------------------
   start_local();  // setup for local time 

   uart_puts(UART_ID, "Choose whether or not to log the data using PB2 (no) or PB1 (yes)\n\n"); 

   SetCmdMode();     // set for commands
   lcd_byte(0x01);   // clear lcd, both lines

   logging = 0;
   domsg(&dmsg8);    // "Log It?  N/Y"
   while(1)
   {
      if( !gpio_get(PB1) )  
      { 
         logging = 1;
         domsg(&dmsg9);    // "Logging it"
         ts_number = 0;    // time stamp number
         uart_puts(UART_ID, "Logging every 5 minutes.\n\n"); 
         break;
      }

      if( !gpio_get(PB2) )  
      { 
         logging = 0;
         domsg(&dmsg10);    // "Not Logging it"
         uart_puts(UART_ID, "Not Logging it\n\n"); 
         break;
      }
      sleep_ms(3000);
   }
//-------------------------
// Create a repeating timer that calls repeating_timer_callback
// If the delay is negative the next callback will be starting from the beginning of the callback
//    Positive starts from the end of the callback
   add_repeating_timer_ms(-4,repeating_timer_callback, NULL, &timer);  // 4 ms
   do_bme280();  // This is the main loop, does not come back

} // end main
//================================================================
//================================================================
void echo_gps_msgs()
{  
   SetCmdMode();     // set for commands
   lcd_byte(0x01);   // clear lcd, both lines
   domsg(&dmsg11);   // Echo GSP Chars
  
   uart_set_fifo_enabled(UART_ID1, true);  // Now
   uart_puts(UART_ID, "\nEcho GPS uart. Press PB3 to stall\n\n");
   while(1)
   {
      // Get GPS uart
      if(  uart_is_readable(UART_ID1)  &&  gpio_get(PB3)  )    // PB3 stalls it while pressed
      { 
         uart_char = uart_getc(UART_ID1); 
         uart_putc(UART_ID, uart_char);
      }
   }
}
//================================================
void blink_led(uint16_t time, uint16_t times)
{
    unsigned int n;

    for(n=0; n<times; n++)
    { 
       gpio_put(LED1,1);
       sleep_ms(time);
       gpio_put(LED1,0);
       sleep_ms(time);
    }
}
//=============================================
void display_utc_lcd()   // utc time
{
   SetCmdMode();     // set for commands
   lcd_byte(0x01);   // clear lcd, both lines

   lcd_wr_instr(0x80);        // position of time
   lcd_wr_data(time.utc_gps_hours[0]);
   lcd_wr_data(time.utc_gps_hours[1]);

   lcd_wr_data(':');
   lcd_wr_data(time.utc_gps_mins[0]);
   lcd_wr_data(time.utc_gps_mins[1]);

   lcd_wr_data(':');
   lcd_wr_data(time.utc_gps_secs[0]);
   lcd_wr_data(time.utc_gps_secs[1]);

   lcd_wr_instr(0xC0);        // position of date
   lcd_wr_data(time.utc_gps_month[0]);
   lcd_wr_data(time.utc_gps_month[1]);

   lcd_wr_data('/');
   lcd_wr_data(time.utc_gps_date[0]);
   lcd_wr_data(time.utc_gps_date[1]);

   lcd_wr_data('/');
   lcd_wr_data(time.utc_gps_year[0]);
   lcd_wr_data(time.utc_gps_year[1]);
   lcd_wr_data(time.utc_gps_year[2]);
   lcd_wr_data(time.utc_gps_year[3]);

   domsg(&dmsg15);  // "UTC"
}
//==================================================
// Testing how exponentiation works here
void test_expo()
{
// test exponentiation
   // double 	8 bytes 	 +/- 1.7e +/- 308 (~15 digits)
   double x, y;
   double result;
   x = 3, y = 2;
   result = pow(x, y);
   int32_t cval = result;

   strcpy(line, "\n");
   uint32_hex_str(cval, nstr);   // uint32 to ASCII string as 8 hex digits
   strcat(line, nstr);
   strcat(line,"\n");
   uart_puts(UART_ID, line);

   // float   4 	+/- 3.4e +/- 38 (~7 digits)
   float x2, y2;
   float result2;
   x2 = 3.0, y2 = 2.0;
   result2 = powf(x2, y2);
   cval = result2;

   //strcpy(line, "\n\n");
   uint32_hex_str(cval, nstr);   // uint32 to ASCII string as 8 hex digits  
   strcpy(line, nstr);
   strcat(line,"\n");
   uart_puts(UART_ID, line);

   x2 = 3.5, y2 = 2.6;
   result2 = powf(x2, y2);
   cval = result2;

   //strcpy(line, "\n\n");
   uint32_hex_str(cval, nstr);   // uint32 to ASCII string as 8 hex digits   
   strcpy(line, nstr);
   strcat(line,"\n");
   uart_puts(UART_ID, line);

   uint32_bcd_str(cval, nstr);   // uint32_t to ASCII string, 8 digits   
   strcpy(line, nstr);
   strcat(line,"\n");
   uart_puts(UART_ID, line);

   cval = result2*1000;
   uint32_bcd_str(cval, nstr);   // uint32_t to ASCII string, 8 digits  
   strcat(line, nstr);
   strcat(line,"\n\n");
   uart_puts(UART_ID, line);

   //LED_break();
}
//==============================
/* Calculate Day of the Week
1. Take the last two digits of the year
2. Add to that 1/4 of these 2 digits (discard any remainder)
3. Add to that the day of the month and the Month Key number for that month
4. From 2000 to 2009 subtract 1 before dividing
5. Divide by 7 and take the remainder. If the remainer is 0 the day is Saturday (7)
*/
uint8_t do_day_of_week(uint8_t month, uint8_t date, uint8_t year)
{
   unsigned char temp;
   
   temp = year + (year>>2);

   if( (year % 4) == 0) 
      { temp = temp + date + month_key_leap[month-1]; } 
   else 
      { temp = temp + date + month_key[month-1];}  // in the table January is 0

   day_of_week = (temp-1) % 7;    // global
   if(day_of_week == 0) {day_of_week = 7;}

   return(day_of_week);
}
//=======================
void do_switch(unsigned char bit, struct myswitch *sptr)    // Bit, switch structure
{
// Debounce and process the switch
// sptr points to the switch's data block
    sptr->debounce  = sptr->debounce <<1;    // shifts in a zero

    if( gpio_get(bit) )        // a zero has already been shifted in
    {
       sptr->debounce++ ;      // so make LSB a one
    }

// Debounced state is either 0x00 or 0xFF
    if(sptr->debounce == 0x00)  {sptr->state = 0x00;}
    if(sptr->debounce == 0xFF)  {sptr->state = 0x01;}

// process the switch
    if(sptr->state != sptr->old)     // we have a change in switch state
    {
       if( (sptr->old == 0x01) && (sptr->state == 0x00) )  // we have a switch press
       {
          sptr->press = 0x01;
       }

       if( (sptr->old == 0x00) && (sptr->state== 0x01) )  // we have a switch release
       {
          sptr->release = 0x01;
       }

       sptr->old = sptr->state;     // the new state is now the old state
    }

// We now know if a switch has been pressed and if a switch has been released
}
//==========================================
// Send Sats in View to LCD
void lcd_satinv()
{
   uint16_t n;

   SetCmdMode();      // set for commands
   lcd_byte(0x01);    // clear screen

   domsg(&dmsg12);         // "Sats in View"
   lcd_wr_instr(0x8D);     // position
   for(n=0; n<4; n++)
   {
      if(satinv[n] == 0)  { break; }   // NULL character
      lcd_wr_data(satinv[n]);
   }
   sleep_ms(2000);
}
//==========================================
// Wait until there are at least 5 satellites in view
void wait_sats()    
{
   uint16_t flag, m, n;

   SetCmdMode();     // set for commands
   lcd_byte(0x01);   // clear lcd, both lines
   domsg(&dmsg12);    // "# Sats in View"

   flag = 1;
   while(flag)
   {
      if(gga_flag)   // latitude, longitude, and altitude; #sats
      { 
           show_gga_buffer(); 

           // Decode only #sats from GGA   
           // get past 7 commas
           n = 0;
           for(m=0; m<7; m++)   
           {
               for(n=n+1; n<64; n++)  // get past a comma
               { 
                 if(gga_buffer[n] == ',')   { break;}
               }
           }  
           n++;   // Get past the last ','             
           // There might be only 1 character
           sats_in_view[0] = gga_buffer[n++];
           sats_in_view[1] = gga_buffer[n];
           if(sats_in_view[1] == ',')   { sats_in_view[1] = 0; }
           else {sats_in_view[2] = 0;}  // NULL to end the string 
   
           gga_flag = 0;
      }

      // stay in this mode here until there are at least 5 satellites in view
      // convert ASCII BCD to binary
      satinv_hex =  ( (sats_in_view[0] - 0x30) * 10) + (sats_in_view[1] - 0x30);

      lcd_wr_instr(0x8E);     // position of # Sats in View data
      lcd_wr_data(sats_in_view[0]);
      lcd_wr_data(sats_in_view[1]);

      if(satinv_hex >= 5)  { flag = 0; }

   }  // end of wait for sats >= 5

// Get leap seconds (in PUBX 04)
   domsg(&dmsg16);    // "Leap Secs"
   sleep_ms(1000);

   flag = 1;
   while(flag) 
   { 
      if(pubx_flag)
      {     
         if( (pubx_buffer[6] == '0') && (pubx_buffer[7] == '4') )    // PUBX 4   Leap Seconds
         {
            show_pubx_buffer();
            flag = 0;
         }
         pubx_flag = 0;
      }
   }

   n = 0;
   for(m = 0; m<6; m++)  // get past 6 commas
   {
      for(n=n+1; n<100; n++)  
        { if(pubx_buffer[n] == ',') { break; } }    // get past comma
   }
         
   n++;
   leap_seconds[0] = pubx_buffer[n++];
   leap_seconds[1] = pubx_buffer[n++];
   leap_seconds[2] = pubx_buffer[n++];
   leap_seconds[3] = 0;  // NULL

   switch(leap_seconds[2])
   {
   case ',':                  // Normal, from GPS
      leap_seconds[2] = 0;    // NULL
      lcd_wr_instr(0xCD);     // position of Leap Secs
      lcd_wr_data(' ');
      lcd_wr_data(leap_seconds[0]);
      lcd_wr_data(leap_seconds[1]);
      break;

   case 'D':                  // from firmware
      lcd_wr_instr(0xCD);     // position of Leap Secs
      lcd_wr_data(leap_seconds[0]);
      lcd_wr_data(leap_seconds[1]);
      lcd_wr_data(leap_seconds[2]);  // should be 'D'
      break;

   default: 
      lcd_wr_instr(0xCD);     // position of Leap Secs
      lcd_wr_data('-');
      lcd_wr_data('-');
      lcd_wr_data('-');
   }

   strcpy(line, "Leap_Seconds: ");
   strcat(line, leap_seconds);
   strcat(line, "\n\n");
   uart_puts(UART_ID, line);

// Clear "Sats In View" Message
//      lcd_wr_instr(0x80);     // position of # Sats in View data
//      for(n=0; n<12; n++)  { lcd_wr_data(' '); }

// Clear "Leap Secs" Message
//      lcd_wr_instr(0x80);     // position of # Sats in View data
//      for(n=0; n<9; n++)  { lcd_wr_data(' '); }
   
}
//============================================
// Display all of it
void show_alltime(void)
{

// GPS strings
   strcpy(line, "utc hours: ");
   memcpy(nstr, time.utc_gps_hours, 3);
   strcat(line, nstr);
   strcat(line, "   ");

   strcat(line, "utc minutes: ");
   memcpy(nstr, time.utc_gps_mins, 3);
   strcat(line, nstr);
   strcat(line, "   ");

   strcat(line, "utc seconds: ");
   memcpy(nstr, time.utc_gps_secs, 3);
   strcat(line, nstr);
   strcat(line, "   ");

   strcat(line, "utc month: ");
   memcpy(nstr, time.utc_gps_month, 3);
   strcat(line, nstr);
   strcat(line, "   ");

   strcat(line, "utc date: ");
   memcpy(nstr, time.utc_gps_date, 3);
   strcat(line, nstr);
   strcat(line, "   ");

   strcat(line, "utc year: ");
   memcpy(nstr, time.utc_gps_year, 5);
   strcat(line, nstr);
   strcat(line, "\n");

   uart_puts(UART_ID, line);

// GPS BCD
   strcpy(line, "utc hours (bcd): ");
   uint8_hex_str(time.utc_bcd_hours, nstr);  // unsigned char to ASCII string as 2 hex digits
   strcat(line, nstr);
   strcat(line, "   ");

   strcat(line, "utc minutes (bcd): ");
   uint8_hex_str(time.utc_bcd_mins, nstr);  // unsigned char to ASCII string as 2 hex digits
   strcat(line, nstr);
   strcat(line, "   ");

   strcat(line, "utc seconds (bcd): ");
   uint8_hex_str(time.utc_bcd_secs, nstr);  // unsigned char to ASCII string as 2 hex digits
   strcat(line, nstr);
   strcat(line, "   ");

   strcat(line, "utc month (bcd): ");
   uint8_hex_str(time.utc_bcd_month, nstr);  // unsigned char to ASCII string as 2 hex digits
   strcat(line, nstr);
   strcat(line, "   ");

   strcat(line, "utc date (bcd): ");
   uint8_hex_str(time.utc_bcd_date, nstr);  // unsigned char to ASCII string as 2 hex digits
   strcat(line, nstr);
   strcat(line, "   ");

   strcat(line, "utc year (bcd): ");
   uint16_hex_str(time.utc_bcd_year, nstr);  // uint16 to ASCII string as 4 hex digits
   strcat(line, nstr);
   strcat(line, "\n");

   uart_puts(UART_ID, line);

// GPS Hex (bin)
   strcpy(line, "utc hours (hex): ");
   uint8_bcd_str(time.utc_hex_hours, nstr);   // unsigned char to ASCII string, 3 digits 
   strcat(line, nstr);
   strcat(line, "   ");

   strcat(line, "utc minutes (hex): ");
   uint8_bcd_str(time.utc_hex_mins, nstr);   // unsigned char to ASCII string, 3 digits
   strcat(line, nstr);
   strcat(line, "   ");

   strcat(line, "utc seconds (hex): ");
   uint8_bcd_str(time.utc_hex_secs, nstr);  // unsigned char to ASCII string, 3 digits
   strcat(line, nstr);
   strcat(line, "   ");

   strcat(line, "utc month (hex): ");
   uint8_bcd_str(time.utc_hex_month, nstr);  // unsigned char to ASCII string, 3 digits
   strcat(line, nstr);
   strcat(line, "   ");

   strcat(line, "utc date (hex): ");
   uint8_bcd_str(time.utc_hex_date, nstr);  // unsigned char to ASCII string, 3 digits
   strcat(line, nstr);
   strcat(line, "   ");

   strcat(line, "utc year (hex): ");
   uint16_bcd_str(time.utc_hex_year, nstr);   // uint16 to ASCII string, 5 digits
   strcat(line, nstr);
   strcat(line, "\n\n");

   uart_puts(UART_ID, line);
}
//=================================
void display_utc_uart()   // utc time
{
// GPS strings
   strcpy(line, "UTC  ");

   strcat(line, time.utc_gps_hours);
   strcat(line, ":");

   strcat(line, time.utc_gps_mins);
   strcat(line, ":");

   strcat(line, time.utc_gps_secs);
   strcat(line, "   ");

   strcat(line, time.utc_gps_month);
   strcat(line, "/");

   strcat(line, time.utc_gps_date);
   strcat(line, "/");
   
   strcat(line, time.utc_gps_year);
   strcat(line, "\n\n");

   uart_puts(UART_ID, line);
}
//==================================================
void show_zda_buffer()
{
   strcpy(line,zda_buffer);
//   strcat(line,"\n");
   uart_puts(UART_ID, line); 
}
//-------------------------------
void show_gga_buffer()
{
   strcpy(line,gga_buffer);
//   strcat(line,"\n");
   uart_puts(UART_ID, line); 
}
//-------------------------------
void show_pubx_buffer()
{
   strcpy(line,pubx_buffer);
//   strcat(line,"\n");
   uart_puts(UART_ID, line); 
}
//-------------------------------
void show_gps_buffer()
{
   strcpy(line,"GPS Buffer:  ");
   strcat(line,gps_buffer);
   strcat(line,"\n");
   uart_puts(UART_ID, line); 
}
//===============================
// Decode GPS time and date from ZDA, GPS data is ASCII
// Example $GPZDA,130652.0000,04,11,2024,00,00*65
// Will be: ,130652.0000,04,11,2024,00,00*65
void decode_ZDA(uint16_t send)
{
   uint16_t n;

  for(n=0; n<100; n++)  
      { if(zda_buffer[n] == ',') { break; } }    // get past comma

   n++;
   time.utc_gps_hours[0] = zda_buffer[n++];
   time.utc_gps_hours[1] = zda_buffer[n++];
   time.utc_gps_hours[2] = 0;  // NULL

   time.utc_gps_mins[0] = zda_buffer[n++];
   time.utc_gps_mins[1] = zda_buffer[n++];
   time.utc_gps_mins[2] = 0;

   time.utc_gps_secs[0] = zda_buffer[n++];
   time.utc_gps_secs[1] = zda_buffer[n++];
   time.utc_gps_secs[2] = 0;

   for(n=n+1; n<100; n++)  
      { if(gga_buffer[n] == ',') { break; } }    // get past comma

   n++;
   time.utc_gps_date[0] = zda_buffer[n++];
   time.utc_gps_date[1] = zda_buffer[n++];
   time.utc_gps_date[2] = 0;
   n++;

   time.utc_gps_month[0] = zda_buffer[n++];
   time.utc_gps_month[1] = zda_buffer[n++];
   time.utc_gps_month[2] = 0;
   n++;

   time.utc_gps_year[0] = zda_buffer[n++];
   time.utc_gps_year[1] = zda_buffer[n++];
   time.utc_gps_year[2] = zda_buffer[n++];
   time.utc_gps_year[3] = zda_buffer[n];
   time.utc_gps_year[4] = 0;

// Convert ASCII to bcd 
   time.utc_bcd_hours = ((time.utc_gps_hours[0] - 0x30)<<4) + (time.utc_gps_hours[1] - 0x30);
   time.utc_bcd_mins  = ((time.utc_gps_mins[0] - 0x30)<<4) + (time.utc_gps_mins[1] - 0x30);
   time.utc_bcd_secs  =  ((time.utc_gps_secs[0] - 0x30)<<4) + (time.utc_gps_secs[1] - 0x30);
   time.utc_bcd_date  =  ((time.utc_gps_date[0] - 0x30)<<4) + (time.utc_gps_date[1] - 0x30);
   time.utc_bcd_month = ((time.utc_gps_month[0] - 0x30)<<4) + (time.utc_gps_month[1] - 0x30);
   time.utc_bcd_year  = ((time.utc_gps_year[0] - 0x30)<<12) + ((time.utc_gps_year[1] - 0x30)<<8) + 
      ((time.utc_gps_year[2] - 0x30)<<4) + (time.utc_gps_year[3] - 0x30);

// Convert ASCII to hex 
   time.utc_hex_hours = ((time.utc_gps_hours[0] - 0x30)*10) + (time.utc_gps_hours[1] - 0x30);
   time.utc_hex_mins  = ((time.utc_gps_mins[0] - 0x30)*10) + (time.utc_gps_mins[1] - 0x30);
   time.utc_hex_secs  =  ((time.utc_gps_secs[0] - 0x30)*10) + (time.utc_gps_secs[1] - 0x30);
   time.utc_hex_date  =  ((time.utc_gps_date[0] - 0x30)*10) + (time.utc_gps_date[1] - 0x30);
   time.utc_hex_month = ((time.utc_gps_month[0] - 0x30)*10) + (time.utc_gps_month[1] - 0x30);
   time.utc_hex_year  = ((time.utc_gps_year[0] - 0x30)*1000) + ((time.utc_gps_year[1] - 0x30)*100) + 
      ((time.utc_gps_year[2] - 0x30)*10) + (time.utc_gps_year[3] - 0x30);

   time.utc_dow = do_day_of_week( (time.utc_hex_month), (time.utc_hex_date), (time.utc_hex_year - 2000) );  // only last 2 digits of the year
// result is 'uint16 day_of_week' Sunday = day 1 to Saturday = day 7;
// For testing
/*
   do_day_of_week(8,4,24);
   do_day_of_week(8,30,24);
   do_day_of_week(9,21,24);
   do_day_of_week(10,30,24);
   do_day_of_week(11,30,24);
   do_day_of_week(12,30,24);
*/
//   uart_puts(UART_ID, line);   // send it out
//LED_break();

//   time12();  // Get the 12 hour time, makes time.local_hex_hours12 and time.pmflag

   if(send)  { show_alltime(); }
}  // end of decode time and date
//============================================
void decode_GGA(uint16_t send)
{
   uint16_t m, n;
   unsigned char temp;

// get to latitude (skip two commas)

   for(n=0; n<100; n++)  
      { if(gga_buffer[n] == ',') { break; } }    // get past comma

   for(n=n+1; n<100; n++)  
      { if(gga_buffer[n] == ',')  { break; }  }  // get past comma

   m = 0;
   for(n=n+1; n<100; n++)  // get latitude
   {
      temp = gga_buffer[n];
      if(temp == ',')  { break;}
      else   { latitude[m++] = temp; }
   }
 
   latitude[m++] = ' ';               // Space
   latitude[m++] = gga_buffer[n+1];  // N or S
   latitude[m] = 0;  // NULL

   m = 0;
   for(n=n+3; n<128; n++)  // get longitude
   {
      temp = gga_buffer[n];
      if(temp == ',')  { break;}
      else             {longitude[m++] = temp;}
   }
   longitude[m++] = ' ';                 // Space
   longitude[m++] = gga_buffer[n+1];    // W or E
   longitude[m] = 0;  // NULL

  for(n=n+1; n<100; n++)  
      { if(gga_buffer[n] == ',')  { break; }  }  // get past comma

  for(n=n+1; n<100; n++)  
      { if(gga_buffer[n] == ',')  { break; }  }  // get past comma

   n++;
   sats_in_view[0] = gga_buffer[n++];
   sats_in_view[1] = gga_buffer[n++];
   sats_in_view[2] = 0;  // NULL

   for(n=n+1; n<100; n++)  
      { if(gga_buffer[n] == ',')  { break; }  }  // get past comma

   m = 0;
   for(n=n+1; n<100; n++)  // Altitude
   {
      temp = gga_buffer[n];
      if(temp == ',')  { break;}
      else             {altitude[m++] = temp;}
   }

   altitude[m++] = ' ';
   altitude[m++] = gga_buffer[n+1];  // M
   altitude[m] = 0;   // NULL
   // -------------------
   // Make formatted strings
   // Latitude string
   latitude_str[0] = latitude[0];
   latitude_str[1] = latitude[1];
   latitude_str[2] = 0;
   strcat(latitude_str, " Deg  ");

   m = 8;
   for(n=2; n<10; n++)   { latitude_str[m++] = latitude[n]; }
   latitude_str[m] = 0;  // NULL

   strcat(latitude_str, " Mins ");

   latitude_str[m+6] = latitude[11];  // N or S
   latitude_str[m+7] = 0;             // NULL
   strcpy(gps_latitude_str, latitude_str);
   // -------------------
   // Longitude string
   longitude_str[0] = longitude[0];
   longitude_str[1] = longitude[1];
   longitude_str[2] = longitude[2];
   longitude_str[3] = 0;
   strcat(longitude_str, " Deg  ");

   m = 9; 
   for(n=3; n<11; n++)   { longitude_str[m++] = longitude[n]; }
   longitude_str[m] = 0;  // NULL

   strcat(longitude_str, " Mins ");

   longitude_str[m+6] = longitude[12];  // W or E
   longitude_str[m+7] = 0;              // NULL
   strcpy(gps_longitude_str, longitude_str);
   // -------------------
   // Altitude string
   strcpy(altitude_str, altitude);   // ASCII Meters
   strcat(altitude_str, "    ");

   float alt = (altitude[0]- 0x30) * 1000.;  // convert to binary so we can convert it to feet
   alt = alt + (altitude[1] - 0x30) * 100;
   alt = alt + (altitude[2] - 0x30) * 10;
   alt = alt + (altitude[3] - 0x30);

   float k = 3.2808;
   float feet = alt * k;
   uint16_bcd_str(alt, nstr);    // uint16 to ASCII string, 5 digits
                 
   uint16_bcd_str(feet, nstr);   // uint16 to ASCII string, 5 digits, ASCII feet
   strcat(altitude_str, nstr);       
   strcat(altitude_str, " Feet");
   //--------------------------------------
   // Make the string for sending it out
   strcpy(line, "Latitude: ");
   strcat(line, latitude_str);
   strcat(line, "    ");

   strcat(line, "Longitude: ");
   strcat(line, longitude_str);
   strcat(line, "    ");

   strcat(line, "GPS Altitude: "); 
   strcat(line, altitude_str);
   strcpy(gps_altitude_str, altitude_str);  // global
   strcat(line,"\n");

   strcat(line,"Sats in View: ");
   strcat(line,sats_in_view);
   strcat(line,"\n\n");

   lcd_wr_instr(0x8E);     // position of # Sats in View data
   lcd_wr_data(sats_in_view[0]);
   lcd_wr_data(sats_in_view[1]);

   if(send) { uart_puts(UART_ID, line);}
//   LED_break();
   return;
}
//=============================================
void decode_PUBX(uint16_t send)
{
   uint16_t m, n;
   char temp;
   char data[16];

// Is it PUBX 0 or PUBX 4?
   if( (pubx_buffer[6] == '0') && (pubx_buffer[7] == '0') )    // PUBX 0 For ellipsoid altitude 
   {
      n = 0;
      for(m = 0; m<7; m++)  // get past 7 commas
      {
         for(n=n+1; n<100; n++)  
            { if(pubx_buffer[n] == ',') { break; } }    // get past comma
      }

      m = 0;
      for(n=n+1;n<100; n++)
      {
         temp  = pubx_buffer[n];
         if(temp == ',')   { break;}
         data[m++] = pubx_buffer[n];    // will become altitude_ellipsoid[]
      }

      data[m++] = 0;    // NULL
      strcpy(line, "Ellipsoid Altitude: ");
      strcat(line, data);
      strcat(line, " M\n\n");
      if(send)   { uart_puts(UART_ID, line); }
      return;
   }

   if( (pubx_buffer[6] == '0') && (pubx_buffer[7] == '4') )    // PUBX 4   Leap Seconds
   {
      n = 0;
      for(m = 0; m<6; m++)  // get past 6 commas
      {
         for(n=n+1; n<100; n++)  
            { if(pubx_buffer[n] == ',') { break; } }    // get past comma
      }
      
      n++;
   
      leap_seconds[0] = pubx_buffer[n++];
      leap_seconds[1] = pubx_buffer[n++];
      leap_seconds[2] = pubx_buffer[n++];
      leap_seconds[3] = 0;  // NULL

      switch(leap_seconds[2])
      {
      case ',':                  // Normal, from GPS
         leap_seconds[2] = 0;    // NULL
         lcd_wr_instr(0xCD);     // position of Leap Secs
         lcd_wr_data(' ');
         lcd_wr_data(leap_seconds[0]);
         lcd_wr_data(leap_seconds[1]);
         break;

      case 'D':                  // Default, from firmware
         lcd_wr_instr(0xCD);     // position of Leap Secs
         lcd_wr_data(leap_seconds[0]);
         lcd_wr_data(leap_seconds[1]);
         lcd_wr_data(leap_seconds[2]);  // should be 'D'
         break;

      default:                   // Error
         leap_seconds[0] = '-' ;
         leap_seconds[1] = '-' ;
         leap_seconds[2] = '-' ;
         leap_seconds[3] = 0;
      }

      strcpy(line, "Leap_Seconds: ");
      strcat(line, leap_seconds);
      strcat(line, "\n\n");
      if(send)   { uart_puts(UART_ID, line); }

      return;
   }
}
//==================================================
void test_messages()
{
   SetCmdMode();     // set for commands
   lcd_byte(0x01);   // clear lcd, both lines
   domsg(&dmsg17);   // "Echo GPS Msgs"

   sleep_ms(2000);
   SetCmdMode();     // set for commands
   lcd_byte(0x01);   // clear lcd, both lines

   while(1)
   {

// For testing
//    if(gps_flag) 
//    { 
//         strcpy(line, "GPS Buffer  ");
//         strcat(line, gps_buffer);
//         strcat(line, "\n");
//         uart_puts(UART_ID,line);
//         display_gps_buffer(1); 
 
      // Do this first
      if(pubx_flag)           // PUBX 0 = Ellipsoid Altitude. PUBX 4 = Leap Seconds
      { 
         show_pubx_buffer(); 
         decode_PUBX(1);
         pubx_flag = 0;
      }

      if(zda_flag)           // time and date, UTC corrected for leap seconds
      {
         show_zda_buffer(); 
         decode_ZDA(1);
         display_utc_lcd();
         zda_flag = 0;
      }
 
      if(gga_flag)           // latitude, longitude, and altitude; #sats
      { 
         show_gga_buffer(); 
         decode_GGA(1);
         gga_flag = 0;
      }

      gps_flag = 0;
      while( !gpio_get(PB3) );   // stall while PB3 is pressed
      sleep_ms(200);  // This seems to be needed
 
   }  // end while

}
//==========================================
// Just show #sats and Leap Seconds while PB1 is pressed
void show_sats_leap()
{
   uint16_t flag, m, n;

   SetCmdMode();       // set for commands
   lcd_byte(0x01);     // clear lcd, both lines
   domsg(&dmsg12);     // "# Sats in View"
   domsg(&dmsg16);     // "Leap Secs"

   flag = 1;
   while(flag)
   {
      if(gga_flag)   // latitude, longitude, and altitude; #sats
      { 
           show_gga_buffer(); 
           // Decode only #sats from GGA   
           // get past 7 commas
           n = 0;
           for(m=0; m<7; m++)   
           {
               for(n=n+1; n<64; n++)  // get past a comma
               { 
                 if(gga_buffer[n] == ',')   { break;}
               }
           }  
           n++;   // Get past the last ','             
           // There might be only 1 character
           sats_in_view[0] = gga_buffer[n++];
           sats_in_view[1] = gga_buffer[n];
           if(sats_in_view[1] == ',')   { sats_in_view[1] = 0; }
           else {sats_in_view[2] = 0;}  // NULL to end the string 
   
           gga_flag = 0;

           lcd_wr_instr(0x8E);     // position of # Sats in View data
           lcd_wr_data(sats_in_view[0]);
           lcd_wr_data(sats_in_view[1]);
      }  // end GGA message

// Get leap seconds (in PUBX 04)
//      sleep_ms(1000);

      if(pubx_flag)
      {     
         show_pubx_buffer(); 

         if( (pubx_buffer[6] == '0') && (pubx_buffer[7] == '4') )    // PUBX 4   Leap Seconds
         {
            n = 0;
            for(m = 0; m<6; m++)  // get past 6 commas
            {
               for(n=n+1; n<100; n++)  
                 { if(pubx_buffer[n] == ',') { break; } }    // get past comma
            }
         
            n++;
            leap_seconds[0] = pubx_buffer[n++];
            leap_seconds[1] = pubx_buffer[n++];
            leap_seconds[2] = pubx_buffer[n++];
            leap_seconds[3] = 0;  // NULL

            switch(leap_seconds[2])
            {
              case ',':                  // Normal, from GPS
                leap_seconds[2] = 0;    // NULL
                lcd_wr_instr(0xCD);     // position of Leap Secs
                lcd_wr_data(' ');
                lcd_wr_data(leap_seconds[0]);
                lcd_wr_data(leap_seconds[1]);
                break;

              case 'D':                  // from firmware
                lcd_wr_instr(0xCD);     // position of Leap Secs
                lcd_wr_data(leap_seconds[0]);
                lcd_wr_data(leap_seconds[1]);
                lcd_wr_data(leap_seconds[2]);  // should be 'D'
                break;

              default: 
                lcd_wr_instr(0xCD);     // position of Leap Secs
                lcd_wr_data('-');
                lcd_wr_data('-');
                lcd_wr_data('-');
            }
         }  // end PUBX 4 message
         pubx_flag = 0;
      }  // end PUBX message

      if(gpio_get(PB1) )    // Go back
      { 
         SetCmdMode();       // set for commands
         lcd_byte(0x01);     // clear lcd, both lines
         return; 
      }   
   
   } // end while
}  
//==========================================
